

#include "meta.h"
#include "lockable.h"

#include "maintenance.h"
#include "expiration.h"
#include "pkgimport.h"
#include "showinfo.h"
#include "aclogger.h"
#include "filereader.h"
#include "acfg.h"
#include "acbuf.h"
#include "sockio.h"

#include <stdio.h>
#include <string.h>
#include <errno.h>
//#include <iostream>

#include <unistd.h>
#include <signal.h>

using namespace MYSTD;

#define MAINT_HTML_DECO "maint.html" 

#ifndef MSG_MORE
#define MSG_MORE 0
#endif

maintenance::maintenance(int fd) :
	m_bErrAbort(false),
	m_reportFD(fd),
	m_szDecoFile(NULL)
{
}

maintenance::~maintenance()
{
}

/// @brief Sends a HTTP chunk with string contents
/// To terminate the transfer, add an empty string with force flag

void maintenance::SendMsg(const string & x) 
{
    SendMsg(x.data(), x.length());
}

inline bool maintenance::_SendHelper(const char *data, size_t len, int flags)
{
	while(len>0)
	{
		int r=send(m_reportFD, data, len, flags);
		if(r<0)
		{
			if(errno==EINTR || errno==EAGAIN)
				r=0;
			else
				return false;
		}
		
		data+=r;
		len-=r;
	}
	return true;
}

void maintenance::SendMsg(const char *data, size_t len)
{
	if(!data || !len || m_reportFD<0)
		return;
	
	char buf[23];
	int l=sprintf(buf, "%x\r\n", (UINT) len);
	_SendHelper(buf, l, MSG_MORE|MSG_NOSIGNAL);
	_SendHelper(data, len, MSG_MORE|MSG_NOSIGNAL);
	_SendHelper("\r\n", 2, MSG_NOSIGNAL);
}

void maintenance::EndTransfer() 
{
	const char msg[] = "0\r\n\r\n"; 
	_SendHelper(msg, sizeof(msg)-1, MSG_NOSIGNAL);
}


void maintenance::SendDecoration(bool bBegin, const char *httpcode, const char *szDecoFile)
{
	if (bBegin)
	{
		const char hs[]="HTTP/1.1 ";
		const char head[] = " OK\r\nConnection: close\r\n"
			"Transfer-Encoding: chunked\r\nContent-Type: text/html\r\n"
			"\r\n";
		_SendHelper(hs, sizeof(hs)-1, MSG_MORE);
		_SendHelper(httpcode, strlen(httpcode), MSG_MORE);
		_SendHelper(head, sizeof(head)-1, MSG_MORE);
	}
	
	acbuf m_deco;
	char *mark(NULL);
	
	if(szDecoFile && m_deco.initFromFile((acfg::confdir+sPathSep+szDecoFile).c_str()))
		mark=strchr(m_deco.c_str(), '~');
	
	// deco file specified and loaded?
	if(mark)
	{
		if(bBegin)
			m_deco.erase(mark-m_deco.rptr());
		else
			m_deco.drop(mark-m_deco.rptr()+1);
		string sDeco(m_deco.rptr(), m_deco.size());
		SetStyle(sDeco);
		SendMsg(sDeco);
	}	
	else
		SendMsg(bBegin?"<html><body>":"</html></body>");
	
	if(!bBegin)
		EndTransfer();
}

class authbounce : public maintenance
{
public:
	// some NOOPs
	authbounce(int fd) : maintenance(fd)
	{
	}
	~authbounce()
	{
	}
	void Action(const MYSTD::string & src)
	{
	}
	void Run(const string &)
	{			
     const char authmsg[] = "HTTP/1.1 401 Not Authorized\r\nWWW-Authenticate: "
        "Basic realm=\"Apt-Cacher NG administration area\"\r\n"
        "Connection: Close\r\n"
        "Content-Type: text/html\r\nContent-Length:81\r\n\r\n"
        "Not Authorized. Please contact Apt-Cacher NG administrator for further questions.";
		_SendHelper(authmsg, sizeof(authmsg)-1, 0);
	}
	
};

void DispatchAndRunMaintTask(const MYSTD::string &cmd, int conFD, const char * szAuthLine)
{	
	maintenance *pWorker(NULL);
	if(!szAuthLine)
		szAuthLine="";

	// not effective, why? signal(SIGPIPE, SIG_IGN);
	
	MYTRY
	{
		//std::cout << "vgl: " << authLine << " und " << acfg::adminauth << "\n";
		
		// admin actions are passed with GET parameters, appended after ?
		if(cmd.find('?')!=stmiss)
		{
			if( ! acfg::adminauth.empty() && acfg::adminauth!=szAuthLine )
				pWorker = new authbounce(conFD);
			else if(cmd.find("doExpire=")!=stmiss)
				pWorker = new expiration(conFD);
			else if(cmd.find("doImport=")!=stmiss)
				pWorker=new pkgimport(conFD);
		}
		else if (cmd==acfg::reportpage)
			pWorker = new reportgen(conFD);
		else
			pWorker = new showinfo(conFD);

		pWorker->m_bErrAbort=(cmd.find("abortOnErrors=aOe")!=stmiss);
	
		pWorker->Run(cmd);
	}
	MYCATCH(...)
	{ /* whatever */ };
	if(pWorker)
		delete pWorker;
}

void maintenance::SetStyle(string &s)
{
	tStrPos pos=0;

	acbuf buf;
	if( ! buf.initFromFile( (acfg::confdir+sPathSep+"style.css").c_str()) )
		return;
	
	while(stmiss != (pos = s.find("$STYLE", pos)))
	{
		//cerr << "replacing " << s.substr(pos, 6) << " with " << buf.c_str()<<endl;
		s.replace(pos, 6, buf.c_str());
		pos+=buf.size();
	}
}

string maintenance::GetHostname()
{
	
	struct sockaddr_storage ss;
	socklen_t slen = sizeof(ss);
	char hbuf[NI_MAXHOST];
	
	if (0==getsockname(m_reportFD, (struct sockaddr *)&ss, &slen) &&
			0==getnameinfo((struct sockaddr*) &ss, sizeof(ss), hbuf, sizeof(hbuf), 
							NULL, 0, NI_NUMERICHOST))
	{
		return hbuf;
	}
	
	return "IP-of-this-cache-server";

}

